"""
pyCLINE: Python implementation of the CLINE method introduced by Prokop, Billen, Frolov and Gelens (2025).

Modules:
    recovery_methods: Methods for recovering nullclines.
    model: Model definitions and utilities.
    generate_data: Functions for generating synthetic data.
    example: Example usage and demonstrations.
"""

from . import recovery_methods
from . import model
from . import generate_data
from . import example
from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("pyCLINE")
except PackageNotFoundError:
    __version__ = "0.0.0"  # Default when package is not installed

__all__ = ["recovery_methods", "model", "generate_data", "example"]



import unittest
import torch
import pandas as pd
from pyCLINE.recovery_methods.nn_training import (
    NeuralNetworkSetupError,
    NeuralNetworkSetupTrainingError,
    NeuralNetworkDataError,
    configure_FFNN_model,
    train_FFNN_model,
    FFNN
)

class TestNNTraining(unittest.TestCase):

    def setUp(self):
        # Setup common variables for tests
        self.Nin = 2
        self.Nout = 1
        self.Nlayers = 2
        self.Nnodes = 8
        self.lr = 1e-4
        self.optimizer_name = 'Adam'
        self.loss_fn = torch.nn.MSELoss
        self.input_train = pd.DataFrame(torch.randn(100, self.Nin).numpy())
        self.target_train = pd.DataFrame(torch.randn(100, self.Nout).numpy())
        self.input_test = pd.DataFrame(torch.randn(20, self.Nin).numpy())
        self.target_test = pd.DataFrame(torch.randn(20, self.Nout).numpy())
        self.validation_data = (pd.DataFrame(torch.randn(20, self.Nin).numpy()), pd.DataFrame(torch.randn(20, self.Nout).numpy()))

    def test_configure_FFNN_model_errors(self):
        with self.assertRaises(NeuralNetworkSetupError):
            configure_FFNN_model(0, self.Nout, self.Nlayers, self.Nnodes)
        with self.assertRaises(NeuralNetworkSetupError):
            configure_FFNN_model(self.Nin, 0, self.Nlayers, self.Nnodes)
        with self.assertRaises(NeuralNetworkSetupError):
            configure_FFNN_model(self.Nin, self.Nout, 0, self.Nnodes)
        with self.assertRaises(NeuralNetworkSetupError):
            configure_FFNN_model(self.Nin, self.Nout, self.Nlayers, 0)
        with self.assertRaises(NeuralNetworkSetupError):
            configure_FFNN_model(self.Nin, self.Nout, self.Nlayers, self.Nnodes, lr=0)
        with self.assertRaises(NeuralNetworkSetupError):
            configure_FFNN_model(self.Nin, self.Nout, self.Nlayers, self.Nnodes, optimizer_name='')
        with self.assertRaises(NeuralNetworkSetupError):
            configure_FFNN_model(self.Nin, self.Nout, self.Nlayers, self.Nnodes, loss_fn=None)

    def test_train_FFNN_model_data_errors(self):
        model, optimizer, loss_fn = configure_FFNN_model(self.Nin, self.Nout, self.Nlayers, self.Nnodes)

        with self.assertRaises(NeuralNetworkDataError):
            train_FFNN_model(model, optimizer, loss_fn, pd.DataFrame(), self.target_train, self.input_test, self.target_test, self.validation_data)
        with self.assertRaises(NeuralNetworkDataError):
            train_FFNN_model(model, optimizer, loss_fn, self.input_train, pd.DataFrame(), self.input_test, self.target_test, self.validation_data)
        with self.assertRaises(NeuralNetworkDataError):
            train_FFNN_model(model, optimizer, loss_fn, self.input_train, self.target_train, pd.DataFrame(), self.target_test, self.validation_data)
        with self.assertRaises(NeuralNetworkDataError):
            train_FFNN_model(model, optimizer, loss_fn, self.input_train, self.target_train, self.input_test, pd.DataFrame(), self.validation_data)
        with self.assertRaises(NeuralNetworkDataError):
            train_FFNN_model(model, optimizer, loss_fn, self.input_train.iloc[:50], self.target_train, self.input_test, self.target_test, self.validation_data)
        with self.assertRaises(NeuralNetworkDataError):
            train_FFNN_model(model, optimizer, loss_fn, self.input_train, self.target_train, self.input_test.iloc[:10], self.target_test, self.validation_data)
        with self.assertRaises(NeuralNetworkDataError):
            train_FFNN_model(model, optimizer, loss_fn, self.input_train, self.target_train, self.input_test, self.target_test, (pd.DataFrame(), self.validation_data[1]))
        with self.assertRaises(NeuralNetworkDataError):
            train_FFNN_model(model, optimizer, loss_fn, self.input_train, self.target_train, self.input_test, self.target_test, (self.validation_data[0], pd.DataFrame()))
        with self.assertRaises(NeuralNetworkDataError):
            train_FFNN_model(model, optimizer, loss_fn, self.input_train, self.target_train, self.input_test, self.target_test, (self.validation_data[0].iloc[:10], self.validation_data[1]))

    def test_train_FFNN_model_setup_errors(self):
        model, optimizer, loss_fn = configure_FFNN_model(self.Nin, self.Nout, self.Nlayers, self.Nnodes)

        with self.assertRaises(NeuralNetworkSetupTrainingError):
            train_FFNN_model(model, optimizer, loss_fn, self.input_train, self.target_train, self.input_test, self.target_test, self.validation_data, epochs=0)
        with self.assertRaises(NeuralNetworkSetupTrainingError):
            train_FFNN_model(model, optimizer, loss_fn, self.input_train, self.target_train, self.input_test, self.target_test, self.validation_data, batch_size=0)

if __name__ == '__main__':
    unittest.main()


import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from scipy.signal import find_peaks

def normalize_variables(df, vars, time, norm_method='minmax', value_max=1.0, value_min=0.0):
    """
    The function conducts min-max normalization in the range '[0:1]
    of the varibles from the list 'vars' in the dataframe 'df'

    return dataframe containing normalized variables 'df_norm'
    and the dataframe containing normalization coefficients per variable 'df_coef'

    Args:
        df (pandas DataFrame): dataframe containing variables
        vars (list): list of the variables to normalize
        time (str): name of the time column
        value_max (float, optional): Maximal value for min-max normalizaiton. Defaults to 1.0.
        value_min (int, optional): Minimal value for min-max noramlization. Defaults to 0.0.

    Returns:
        df_norm (pandas DataFrame): dataframe containing normalized variables
        df_coef (pandas DataFrame): dataframe containing minimal and maximal values per variable
    """
    # creating a new dataframe for normalized data and copying the time column
    df_norm = pd.DataFrame()
    df_norm[time] = df[time]
    # creating a dataframe to store the normalization coefficients (min and max of the raw variables)
    df_coef = pd.DataFrame(index=['min', 'max'])
    if norm_method != 'None':
        # normalizing variables
        for var in vars:
            if norm_method == 'minmax':
                df_coef[var] = [df[var].min(), df[var].max()]
                
                df_norm['norm {:}'.format(var)] = (df[var] - df_coef[var]['min'])*(value_max-value_min)/(df_coef[var]['max'] - df_coef[var]['min']) + value_min
            if norm_method == 'zscore':
                df_coef[var] = [df[var].mean(), df[var].std()]
                df_norm['norm {:}'.format(var)] = (df[var] - df_coef[var]['mean'])/(df_coef[var]['sd'])

    return df_norm, df_coef

def plot_optimal_thresholding(thresholds, nsamples, varsamples, optimal_threshold, idx, histogram):
    """
    When a system exhibits a non-uniform distribution of data points in the phase space,
    the optimal thresholding can be used to sample the data uniformly.
    This function plots the normalized sample size, normalized sampling standard deviation,
    and the ratio of the sample size to the sampling standard deviation as a function of the threshold.

    Args:
        thresholds (numpy array): array of threshold values
        nsamples (numpy array): array of normalized sample sizes
        varsamples (numpy array): array of normalized sampling standard deviations
        optimal_threshold (int): optimal threshold value
        idx (int): index of the optimal threshold value
        histogram (numpy histogram object): 2D histogram of the phase space
    """  
    fig, axes = plt.subplots(1,2, figsize=(11,5))
    axes[0].plot(thresholds, nsamples, label='norm. sample size')
    axes[0].plot(thresholds, varsamples, label='norm. sampling SD', c='C1')
    axtwinx = axes[0].twinx()
    axtwinx.plot(thresholds, nsamples/varsamples, label='sample size/SD', c='C2')
    axtwinx.scatter(optimal_threshold, (nsamples/varsamples)[idx], c='r', s=50, label='optimal threshold, {:d}'.format(int(optimal_threshold)))

    axes[0].legend()
    axtwinx.legend()
    axes[0].set_xlabel('sampling threshold')
    axes[0].set_ylabel('norm sample size, sampling SD')
    axtwinx.set_ylabel('sample size/SD')

    hthresh = np.copy(histogram)
    hthresh[hthresh>optimal_threshold] = optimal_threshold

    axes[1].imshow(hthresh)

    plt.show()

def compute_optimal_threshold(df, vars, binx, biny, plot_thresholding=True):
    """
    When a system exhibits a non-uniform distribution of data points in the phase space,
    the optimal thresholding can be used to sample the data uniformly.
    This function computes the optimal thresholding value based on the normalized sample size,
    normalized sampling standard deviation, and the ratio of the sample size to the sampling standard deviation.

    Args:
        df (pandas dataframe): dataframe containing variables
        vars (list): list of the variables to generate the optimal threshold
        binx (list): list of bins for the first variable
        biny (list): list of bins for the second variable
        plot_thresholding (bool, optional): If the thresholded phase space between the first and second
                                            variable should be plotted. Defaults to True.

    Returns:
        optimal_threshold (int): optimal threshold value
    """    
    h, _, _ = np.histogram2d(df[vars[0]], df[vars[1]], bins=(binx, biny))

    nthresh = np.arange(np.min(h[h>0]),np.max(h[h>0]),50)
    nsamples = np.zeros(nthresh.shape[0])
    varsamples = np.zeros(nthresh.shape[0])

    for i in range(nthresh.shape[0]):

        hthresh = np.copy(h)
        hthresh[hthresh>nthresh[i]] = nthresh[i]
        nsamples[i] = np.sum(hthresh)
        varsamples[i] = np.std(hthresh[hthresh>0])

    if np.max(nsamples) != np.min(nsamples):
        nsamples = (nsamples - np.min(nsamples))/(np.max(nsamples) - np.min(nsamples)) + 1
        varsamples = (varsamples - np.min(varsamples))/(np.max(varsamples) - np.min(varsamples)) + 1

        idx = np.argmax(nsamples/varsamples)
        optimal_threshold = nthresh[idx]
    else:
        optimal_threshold = np.max(nthresh)

    if plot_thresholding:
        plot_optimal_thresholding(nthresh, nsamples, varsamples, optimal_threshold, idx, h)

    return int(optimal_threshold)

def uniform_sampling(df, threshold, input_vars, binx, biny):
    """
    The function samples the data uniformly in the phase space defined by the input variables
    based on the provided threshold value.

    Args:
        df (pandas dataframe): dataframe containing variables
        threshold (int): threshold value for the uniform sampling
        input_vars (list): list of the input variables
        binx (list): list of bins for the first variable
        biny (list): list of bins for the second variable

    Returns:
        df_uniform (pandas dataframe): dataframe containing uniformly sampled data
    """    
    df_uniform = pd.DataFrame()

    for i in range(0,binx.shape[0]):
        for j in range(0,biny.shape[0]):

            df_subsample = df[(df[input_vars[0]]>binx[i-1]) & (df[input_vars[0]]<binx[i]) &
                              (df[input_vars[1]]>biny[j-1]) & (df[input_vars[1]]<biny[j])].copy()
            if df_subsample.shape[0]>0:
                if df_subsample.shape[0]<=threshold:
                    df_uniform = pd.concat((df_uniform,df_subsample), ignore_index=True)
                else:
                    df_uniform = pd.concat((df_uniform,df_subsample[:threshold]), ignore_index=True)

    return df_uniform

# data preparation
def prepare_data(df, vars, time, tmin=None, tmax=None, scheme='newton_difference', norm_method='minmax',value_min=0.0,  value_max=1.0, normalize=True):    
    """
    The function prepares the raw time series of the system variables for feeding into ML model. Preparation includes the following steps: (i) data slicing in the indicated range [tmin:tmax], [:tmax], or [tmin:] (optional). If tmin and tmax are not provided, full data are processed (ii) min-max normalization of the variables from the list 'vars'(iii) computing a delayed [t-1] variable for each variable from the list.
    The function returns a prepared dataframe 'df_prepared' and the dataframe containing normalization coefficients per variable 'df_coef'.

    Args:
        df (pandas dataframe): dataframe containing variables
        vars (list): list of the variables to normalize
        time (str): name of the time column
        tmin (float, optional): Minimal time of the time slice of data. Defaults to None.
        tmax (float, optional): Maximal time of the time slice of data. Defaults to None.
        scheme (str, optional): Scheme for computing second input variable. Defaults to 'newton_difference'.
        norm_method (str, optional): Normalization method. Defaults to 'minmax'.
        value_min (int, optional): Minimal value for minmax normalization. Defaults to 0.0.
        value_max (int, optional): Maximal value for minmax normalization. Defaults to 1.0.
        normalize (bool, optional): If data should be normalized. Defaults to True.

    Returns:
        df_prepared (pandas dataframe): dataframe containing prepared data for feeding into ML model
        df_coef (pandas dataframe): dataframe containing normalization coefficients (min and max values) per variable
    """   
    
    if not isinstance(df, pd.DataFrame):
        raise ValueError("Input data is not a pandas DataFrame")
    if not isinstance(vars, list):
        raise ValueError("Variables should be provided as a list")
    if not isinstance(time, str):
        raise ValueError("Time column name of pandas DataFrame should be provided as a string")
    if not isinstance(tmin, (int, float)) and tmin is not None:
        raise ValueError("tmin should be a float or an integer, or None")
    if not isinstance(tmax, (int, float)) and tmax is not None:
        raise ValueError("tmax should be a float or an integer, or None")
    if not isinstance(scheme, str):
        raise ValueError("Scheme should be provided as a string")
    if not isinstance(norm_method, str):
        raise ValueError("Normalization method should be provided as a string")
    if not isinstance(value_min, (int, float)):
        raise ValueError("Minimal value for minmax normalization should be a float or an integer")
    if not isinstance(value_max, (int, float)):
        raise ValueError("Maximal value for minmax normalization should be a float or an integer")
    if not isinstance(normalize, bool):
        raise ValueError("Normalize variable should be a boolean value")
    if not all([var in df.columns for var in vars]):
        raise ValueError("Variables should be present in the provided dataframe")
    if not time in df.columns:
        raise ValueError("Time column should be present in the provided dataframe")
    if scheme not in ['newton_difference', 'two_point', 'five_point', 'derivative']:
        raise ValueError("Unknown scheme: {:}".format(scheme))

    # slice the data in the range [tmin; tmax] if needed
    # if ((tmin is None) or (tmax is None)):
    if tmin is None:
        tmin = df[time].min()
    if tmax is None:
        tmax = df[time].max()
    df_slice = df[(df[time]>=tmin) & (df[time]<=tmax)].copy()

    # min-max normalization of each variable in the range [value_min; value_max]
    if normalize:
        df_norm, df_coef = normalize_variables(df_slice, vars=vars, time=time, norm_method=norm_method,value_max=value_max,value_min=value_min)
    else:
        df_norm = df_slice.copy()
        df_coef = pd.DataFrame(index=['min', 'max'])
        for var in vars:
            df_norm['norm {:}'.format(var)] = df_norm[var]
            df_coef[var] = [df[var].min(), df[var].max()]
    # computing delayed variables
    df_prepared = pd.DataFrame()
    
    if scheme == 'newton_difference':
        first_point = 1
    if scheme == 'two_point':
        first_point = 2   
    if scheme == 'five_point':
        first_point = 4
    if scheme == 'derivative':
        first_point=0
    df_prepared[time] = df_norm[time].iloc[first_point:]
    for var in vars:
        # df_prepared['norm {:}'.format(var)] = df_norm['norm {:}'.format(var)].to_numpy()[first_point:]
        if scheme=='newton_difference':
            df_prepared['norm {:}'.format(var)] = df_norm['norm {:}'.format(var)].to_numpy()[first_point:]
            df_prepared['norm {:}'.format(var)+'[t-1]'] = df_norm['norm {:}'.format(var)].to_numpy()[:-first_point]
        if scheme=='two_point':
            df_prepared['norm {:}'.format(var)] = df_norm['norm {:}'.format(var)].to_numpy()[first_point-1:-first_point+1]
            df_prepared['norm {:}'.format(var)+'[t+1]'] = df_norm['norm {:}'.format(var)].to_numpy()[first_point:]
            df_prepared['norm {:}'.format(var)+'[t-1]'] = df_norm['norm {:}'.format(var)].to_numpy()[first_point-2:-first_point]
        if scheme=='five_point':
            df_prepared['norm {:}'.format(var)] = df_norm['norm {:}'.format(var)].to_numpy()[first_point-2:-first_point+2]
            df_prepared['norm {:}'.format(var)+'[t+2]'] = df_norm['norm {:}'.format(var)].to_numpy()[first_point:]
            df_prepared['norm {:}'.format(var)+'[t+1]'] = df_norm['norm {:}'.format(var)].to_numpy()[first_point-1:-first_point+3]
            df_prepared['norm {:}'.format(var)+'[t-1]'] = df_norm['norm {:}'.format(var)].to_numpy()[first_point-3:-first_point+1]
            df_prepared['norm {:}'.format(var)+'[t-2]'] = df_norm['norm {:}'.format(var)].to_numpy()[:-first_point]
        if scheme=='derivative':
            dt=df[time][1]-df[time][0]
            df_prepared['norm {:}'.format(var)] = df_norm['norm {:}'.format(var)].to_numpy()[first_point:]

            x = df_norm['norm {:}'.format(var)].to_numpy()[first_point:]
            x_dot = np.ones(x.shape[0])*np.NaN
            x_dot_forward =[(x[i+1]-x[i])/dt for i in range(x[:-1].shape[0])]
            
            x_dot_backward =[(x[i]-x[i-1])/dt for i in range(x[1:].shape[0])]

            x_dot_forward = np.array(x_dot_forward)
            x_dot_backward = np.array(x_dot_backward)
            x_dot= (x_dot_forward + x_dot_backward)/2
            # insert np.nan at the first position
            x_dot = np.insert(x_dot, -1, np.nan)
            x_dot[0]=np.nan
            
            df_prepared['d norm{:}'.format(var)+'/dt'] = x_dot

    return df_prepared.dropna(), df_coef

def shuffle_and_split(df, input_vars, target_var, train_frac=0.7, test_frac=0.15, optimal_thresholding=True, plot_thresholding=True):
    """
    The function prepares training, testing, and validation sets from the prepared dataframe
    by random uniform shuffling and splitting the data according to the provided proportions

    the function returns respective training, testing, and validation datasets

    Args:
        df (pandas dataframe): dataframe containing prepared data for feeding into ML model
        input_vars (list): list of the input variables
        target_var (list): list of the target variable(s)
        train_frac (float, optional): fraction of training data. Defaults to 0.7.
        test_frac (float, optional): fraction of testing data. Defaults to 0.15.
        optimal_thresholding (bool, optional): If the optimal thresholding should be used. Defaults to True.
        plot_thresholding (bool, optional): _description_. Defaults to True.

    Returns:
        input_train (pandas dataframe): dataframe containing input variables for training
        target_train (pandas dataframe): dataframe containing target variables for training
        input_test (pandas dataframe): dataframe containing input variables for testing
        target_test (pandas dataframe): dataframe containing target variables for testing
        input_val (pandas dataframe): dataframe containing input variables for validation
        target_val (pandas dataframe): dataframe containing target variables for validation
    """    

    if not isinstance(df, pd.DataFrame):
        raise ValueError("Input data is not a pandas DataFrame")
    if not isinstance(input_vars, list):
        raise ValueError("Input variables should be provided as a list")
    if not isinstance(target_var, list):
        raise ValueError("Target variables should be provided as a list")
    if not isinstance(train_frac, float):
        raise ValueError("Training fraction should be a float")
    if not isinstance(test_frac, float):
        raise ValueError("Testing fraction should be a float")
    if train_frac + test_frac > 1:
        raise ValueError("Training and testing fractions should be less or equal to 1")
    if train_frac< test_frac:
        raise ValueError("Training fraction should be greater than testing fraction")
    if not isinstance(optimal_thresholding, bool):
        raise ValueError("Optimal thresholding should be a boolean value")
    if not isinstance(plot_thresholding, bool):
        raise ValueError("Plot thresholding should be a boolean value")
    if not all([var in df.columns for var in input_vars]):
        raise ValueError("Input variables should be present in the provided dataframe")
    if not all([var in df.columns for var in target_var]):
        raise ValueError("Target variables should be present in the provided dataframe")
    
    # data shuffling
    df_shuffled = df.sample(frac = 1)

    # uniform data sampling in the phase space
    if optimal_thresholding:
        binx = np.linspace(df_shuffled[input_vars[0]].min(), df_shuffled[input_vars[0]].max(), 11)
        biny = np.linspace(df_shuffled[target_var[0]].min(), df_shuffled[target_var[0]].max(), 11)

        if optimal_thresholding:
            optimal_threshold = compute_optimal_threshold(df_shuffled, [input_vars[0], target_var[0]], binx, biny, plot_thresholding)

        df_shuffled = uniform_sampling(df_shuffled, optimal_threshold, [input_vars[0], target_var[0]], binx, biny)
        df_shuffled = df_shuffled.sample(frac = 1)

    # computing the sizes of the training and testing datasets based of the provided fractions
    Ntrain = int(df_shuffled.shape[0]*train_frac)
    Ntest = int(df_shuffled.shape[0]*test_frac)

    # splitting the data
    input_train = df_shuffled[input_vars].iloc[:Ntrain]
    target_train = df_shuffled[target_var].iloc[:Ntrain]

    input_test = df_shuffled[input_vars].iloc[Ntrain:Ntrain+Ntest]
    target_test = df_shuffled[target_var].iloc[Ntrain:Ntrain+Ntest]

    input_val = df_shuffled[input_vars].iloc[Ntrain+Ntest:]
    target_val = df_shuffled[target_var].iloc[Ntrain+Ntest:]

    return input_train, target_train, input_test, target_test, input_val, target_val

def normalize_adjusted(x, df_coef, var, min=-1, max=1):
    return (x - df_coef[var].min())*(max-min) / (df_coef[var].max() - df_coef[var].min())+min


#### General functions
def calculate_period(x_train, t_train):
    """
    Calculate the period of an oscillation.

    Parameters
    ----------
    x_train : np.array
        Time series of data.
    t_train : np.array
        Time of the time series.

    Returns
    -------
    peaks_u : np.array
        Size of peaks detected in the time series.
    peaks_t : np.array
        Time points where peaks occur in the data set.
    period : float
        Period of Time Series.
    peaks[0]: list
        Array of all indicies of local maxima
    
    """
    if len(x_train.shape)<2:
        peaks=find_peaks(x_train)
        peaks_t=t_train[peaks[0]]
        peaks_u=x_train[peaks[0]]
    elif x_train.shape[1]==2 or x_train.shape[1]==3:
        peaks=find_peaks(x_train[:,0])
        peaks_t=t_train[peaks[0]]
        peaks_u=x_train[peaks[0],0]
    elif x_train.shape[1]==5:
        peaks=find_peaks(x_train[:,1])
        peaks_t=t_train[peaks[0]]
        peaks_u=x_train[peaks[0],1]

    period_temp=0
    for i in range(len(peaks_t)-1):
        period_temp=period_temp+(peaks_t[i+1]-peaks_t[i])
    if len(peaks_t)>1:
        subtract=1
    else: subtract=0
    period=period_temp/(len(peaks_t)-subtract)
    return peaks_u, peaks_t, period, peaks[0]



import unittest
import pandas as pd
import numpy as np
from pyCLINE.recovery_methods.data_preparation import prepare_data, shuffle_and_split

class TestDataPreparation(unittest.TestCase):

    def setUp(self):
        # Setup common variables for tests
        self.df = pd.DataFrame({
            'time': np.linspace(0, 10, 100),
            'var1': np.sin(np.linspace(0, 10, 100)),
            'var2': np.cos(np.linspace(0, 10, 100))
        })
        self.vars = ['var1', 'var2']
        self.time = 'time'

    def test_prepare_data(self):
        # Test prepare_data function
        df_prepared, df_coef = prepare_data(self.df, self.vars, self.time)
        self.assertTrue('norm var1' in df_prepared.columns)
        self.assertTrue('norm var2' in df_prepared.columns)
        self.assertTrue('var1' in df_coef.columns)
        self.assertTrue('var2' in df_coef.columns)

    def test_prepare_data_invalid_input(self):
        # Test prepare_data function with invalid inputs
        with self.assertRaises(ValueError):
            prepare_data(self.df, 'var1', self.time)
        with self.assertRaises(ValueError):
            prepare_data(self.df, self.vars, 123)
        with self.assertRaises(ValueError):
            prepare_data(self.df, self.vars, self.time, tmin='invalid')
        with self.assertRaises(ValueError):
            prepare_data(self.df, self.vars, self.time, tmax='invalid')
        with self.assertRaises(ValueError):
            prepare_data(self.df, self.vars, self.time, scheme=123)
        with self.assertRaises(ValueError):
            prepare_data(self.df, self.vars, self.time, norm_method=123)
        with self.assertRaises(ValueError):
            prepare_data(self.df, self.vars, self.time, value_min='invalid')
        with self.assertRaises(ValueError):
            prepare_data(self.df, self.vars, self.time, value_max='invalid')
        with self.assertRaises(ValueError):
            prepare_data(self.df, self.vars, self.time, normalize='invalid')

    def test_shuffle_and_split(self):
        # Test shuffle_and_split function
        input_train, target_train, input_test, target_test, input_val, target_val = shuffle_and_split(
            self.df, self.vars, ['var1'], train_frac=0.7, test_frac=0.15, optimal_thresholding=False
        )
        self.assertIsInstance(input_train, pd.DataFrame)
        self.assertIsInstance(target_train, pd.DataFrame)
        self.assertIsInstance(input_test, pd.DataFrame)
        self.assertIsInstance(target_test, pd.DataFrame)
        self.assertIsInstance(input_val, pd.DataFrame)
        self.assertIsInstance(target_val, pd.DataFrame)

    def test_shuffle_and_split_invalid_input(self):
        # Test shuffle_and_split function with invalid inputs
        with self.assertRaises(ValueError):
            shuffle_and_split(self.df, 'var1', ['var1'])
        with self.assertRaises(ValueError):
            shuffle_and_split(self.df, self.vars, 'var1')
        with self.assertRaises(ValueError):
            shuffle_and_split(self.df, self.vars, ['var1'], train_frac='invalid')
        with self.assertRaises(ValueError):
            shuffle_and_split(self.df, self.vars, ['var1'], test_frac='invalid')
        with self.assertRaises(ValueError):
            shuffle_and_split(self.df, self.vars, ['var1'], train_frac=0.8, test_frac=0.3)
        with self.assertRaises(ValueError):
            shuffle_and_split(self.df, self.vars, ['var1'], train_frac=0.1, test_frac=0.2)
        with self.assertRaises(ValueError):
            shuffle_and_split(self.df, self.vars, ['var1'], optimal_thresholding='invalid')
        with self.assertRaises(ValueError):
            shuffle_and_split(self.df, self.vars, ['var1'], plot_thresholding='invalid')

if __name__ == '__main__':
    unittest.main()


import sys 
import os
from . import generate_data
import pandas as pd
import matplotlib.pyplot as plt
from . import recovery_methods
from . import model
import torch.nn as nn
import numpy
import torch

class ExampleCallable:
    def __call__(self, example_model, plot=True):
        example(example_model, plot)

sys.modules[__name__]= ExampleCallable()

def example(example_model, plot):
    """
    This function runs multiple examples, depending on the choice of the example model. 
    It should be used as a guideline how to run pyCLINE for synthetic data. 
    For using pyCLINE on real data, the user should adapt the data source accordingly.

    Args:
        example_model (str): Selection of model to run the example on. Chose from 'FHN', 'Bicubic', 'GeneExpression', 'DelayOscillator'.
        plot (bool): If True, the function will plot the data and the predictions.

    Raises:
        ValueError: In case no example model string is provided.
    """

    #provide name of model to run example
    if example_model is None:
        raise ValueError("Error: example_model is None: please provide a model name (FHN, Bicubic, GeneExpression, DelayOscillator).")
    
    # check if data already exists
    path='data/synthetic_data/'
    if example_model=='FHN': 
        fname='FHN_eps=0.3_a=0.0.csv'
    else: 
        fname=example_model+'.csv'
    
    print('Running example for model: '+example_model)
    print('Step 1: Load or generate data')
    # generate data if it does not exist
    if os.path.exists(path+fname):
        print('Data already exists: '+path+fname)
        df = pd.read_csv(path+fname)

    else:
        print('No data saved: generating data for model: '+example_model)
        getattr(generate_data, example_model)()
        print('Data generated saved: '+path+fname)
        df = pd.read_csv(path+fname)
    
    # extracting 1 time series and plotting
    if example_model!='DelayOscillator':
        df_sim = df[(df['sim']==1)].copy()
    else:
        df_sim=df.copy()
    df_sim.reset_index(drop=True, inplace=True)

    if example_model=='DelayOscillator':
        tau=10 # 12,20,35
        v_data=df_sim['u'][df_sim['time']>=tau].values
        df_sim=df_sim[df_sim['time']<=df_sim['time'].max() - tau]
        df_sim['v']=v_data

    if plot:
        fig,ax = plt.subplots(1,1,figsize=(5,3))
        ax.plot(df_sim['time'], df_sim['u'], label='u')
        ax.plot(df_sim['time'], df_sim['v'], label='v')
        ax.set_xlabel('Time')
        ax.set_ylabel('Amount in arbs. units')
        plt.show()
    
    # prepare data for training
    # tmin: minimum time point to consider to avoid transient behavior
    # val_min, val_max: min and max values of the min max normalization
    print('Step 2: Prepare data for training')
    if example_model=='FHN':
        tmin, val_min, val_max=3, -1.0, 1.0
    if example_model=='Bicubic':
        tmin, val_min, val_max=10, 0.0, 1.0 
    if example_model=='GeneExpression':
        tmin, val_min, val_max=10, 0.0, 1.0
    if example_model=='DelayOscillator':
        tmin, val_min, val_max=100,0.0,1.0


    # normalize data and create additional derivative or delayed variables
    df_sim, df_coef = recovery_methods.data_preparation.prepare_data(df_sim, vars=['u', 'v'], time='time', tmin=tmin, scheme='derivative', value_min=val_min, value_max=val_max)

    # define input and target variables (can be changed as needed)
    input_vars, target_vars = ['norm u', 'd normu/dt'], ['norm v']

    # schuffle and split training, test and validation data
    # optimal_thresholding: if True, the amount of samples within the phase space is evenly distributed
    input_train, target_train, input_test, target_test, input_val, target_val = recovery_methods.data_preparation.shuffle_and_split(df_sim, input_vars = input_vars,
                                                                                                                    target_var = target_vars,
                                                                                                                    optimal_thresholding=False)
    
    # set up the model
    # Nin: number of input variables
    # Nout: number of output variables
    # Nlayers: number of layers
    # Nnodes: number of nodes per layer
    # summary: if True, the model summary is printed
    # lr: learning rate
    # activation: activation function can be chosen as needed
    print('Step 3: Set up the model')
    nn_model,  optimizer, loss_fn = recovery_methods.nn_training.configure_FFNN_model(Nin=len(input_vars), Nout=len(target_vars),
                                                                                   Nlayers=3, Nnodes=64, summary=True, lr=1e-4,
                                                                                   activation=nn.SiLU)
    
    # train the model
    print('Step 4: Train the model')
    training_loss, val_loss, test_loss, predictions_evolution, lc_predictions = recovery_methods.nn_training.train_FFNN_model(model=nn_model,
                                                                                                                            optimizer=optimizer, loss_fn=loss_fn,
                                                                                                                            input_train=input_train,
                                                                                                                            target_train=target_train,input_test=input_test, 
                                                                                                                            target_test=target_test, 
                                                                                                                            validation_data=(input_val, target_val),
                                                                                                                            epochs=3000, batch_size=64, device='cpu',save_evolution=True, 
                                                                                                                            method='derivative', minimal_value=val_min,maximal_value=val_max)

    # save model and predictions
    print('Step 5: Save the generated predictions and model')
    if not os.path.exists(f'results/{example_model}'):
        os.makedirs(f'results/{example_model}')

    torch.save(nn_model.state_dict(), f'results/{example_model}/model.pth')
    numpy.save(f'results/{example_model}/predictions.npy', predictions_evolution)
    numpy.save(f'results/{example_model}/lc_predictions.npy', lc_predictions)
    numpy.save(f'results/{example_model}/training_loss.npy', training_loss)
    numpy.save(f'results/{example_model}/val_loss.npy', val_loss)
    numpy.save(f'results/{example_model}/test_loss.npy', test_loss)

    print('Example completed: model and predictions saved in results/'+example_model)

    # plot the predictions
    if plot:
        print('Step 6: Plot the predictions')
        fig,ax = plt.subplots(1,1,figsize=(5,3))
        ax.scatter(df_sim['norm u'], df_sim['norm v'], label='GT LC', c='silver')
        ax.scatter(input_train['norm u'], lc_predictions[-1,:,0],
                c='C2', label='Pred. LC', s=2 )
        
        # compute nullcline
        u = numpy.linspace(df_coef['u'].min(), df_coef['u'].max(), predictions_evolution.shape[1])
        sim_model = getattr(model, example_model)()
        gt_nullcline=sim_model.vnull(u)
        
        norm_u=recovery_methods.data_preparation.normalize_adjusted(u, df_coef,'u', 
                                                                    min=val_min, max=val_max)
        norm_gt_nullcline=recovery_methods.data_preparation.normalize_adjusted(gt_nullcline, df_coef,'v', 
                                                                               min=val_min, max=val_max)
        
        ax.plot(norm_u,norm_gt_nullcline, label='GT NC', c='k')
        ax.plot(norm_u,predictions_evolution[-1,:], label='Pred. NC', c='C1')
        ax.set_xlabel('u')
        ax.set_ylabel('v')
        ax.legend()
        plt.show()



"""
recovery_methods: Methods for recovering nullclines.

Modules:
    data_preparation: Functions for preparing data.
    nn_training: Functions for configuring and training neural network models.
"""

from .data_preparation import prepare_data
from .nn_training import configure_FFNN_model, train_FFNN_model

from . import nn_training
from . import data_preparation

__all__ = ['nn_training', 'data_preparation', 'prepare_data', 'configure_FFNN_model', 'train_FFNN_model']


import unittest
import numpy as np
from pyCLINE.generate_data import FHN, Bicubic, GeneExpression, DelayOscillator
from pyCLINE.model import FHN as FHNModel, Bicubic as BicubicModel, GeneExpression as GeneExpressionModel, DelayOscillator as DelayOscillatorModel
import pandas as pd

class TestGenerateData(unittest.TestCase):

    def test_FHN(self):
        # Test FHN data generation
        try:
            FHN(dt=0.1, N=1000, epsilons=[0.3], n_intiaL_conditions=1)
        except Exception as e:
            self.fail(f"FHN data generation failed with exception: {e}")
        
        # Check the shape of the generated data
        fhn = FHNModel([1, 1, 0.3, 0.5, 0.0])
        u0, v0 = np.meshgrid(np.linspace(-1.25, 1.75, 1), np.linspace(-0.75, 1.75, 1))
        x0 = np.array([u0, v0])
        fhn.generate_data(x0, 0.01, 100, check_period=False)
        
        path= 'data/synthetic_data/'
        fname='FHN_eps=0.3_a=0.0.csv'
        data = pd.read_csv(path+fname)

        self.assertEqual(data.shape, (100,4))

    def test_Bicubic(self):
        # Test Bicubic data generation
        try:
            Bicubic(dt=0.1, N=1000, n_intiaL_conditions=1)
        except Exception as e:
            self.fail(f"Bicubic data generation failed with exception: {e}")
        
        # Check the shape of the generated data
        bicubic = BicubicModel([-0.5, 0.5, -1/3])
        u0, v0 = np.meshgrid(np.linspace(-1.25, 1.75, 1), np.linspace(-0.75, 1.75, 1))
        x0 = np.array([u0, v0])
        bicubic.generate_data(x0, 0.01, 100, check_period=False)

        path= 'data/synthetic_data/'
        fname='Bicubic.csv'
        data = pd.read_csv(path+fname)

        self.assertEqual(data.shape, (100,4))

    def test_GeneExpression(self):
        # Test GeneExpression data generation
        try:
            GeneExpression(dt=0.1, N=1000, n_intiaL_conditions=1)
        except Exception as e:
            self.fail(f"GeneExpression data generation failed with exception: {e}")
        
        # Check the shape of the generated data
        gene_expression = GeneExpressionModel([1, 0.05, 1, 0.05, 1, 0.05, 1, 1, 0.1, 2])
        u0, v0 = np.meshgrid(np.linspace(0, 1.75, 1), np.linspace(0, 1.75, 1))
        x0 = np.array([u0, v0])
        gene_expression.generate_data(x0, 0.01, 100, check_period=False)

        path= 'data/synthetic_data/'
        fname='GeneExpression.csv'
        data = pd.read_csv(path+fname)

        self.assertEqual(data.shape, (100,4))

    def test_DelayOscillator(self):
        # Test DelayOscillator data generation
        try:
            DelayOscillator(N=1000)
        except Exception as e:
            self.fail(f"DelayOscillator data generation failed with exception: {e}")
        
        # Check the shape of the generated data
        delay_osci = DelayOscillatorModel([4, 10, 2])
        delay_osci.generate_data(y_0=0, dt=0.4, t_max=400, check_period=False)
        path= 'data/synthetic_data/'
        fname='DelayOscillator.csv'
        data = pd.read_csv(path+fname) 
        self.assertEqual(data.shape, (int(400/0.4),2))

    # def test_invalid_parameters(self):
    #     # Test invalid parameters for FHN
    #     with self.assertRaises(ValueError):
    #         FHN(dt=-0.1, N=1000, epsilons=[0.3], n_intiaL_conditions=1)
    #     with self.assertRaises(ValueError):
    #         FHN(dt=0.1, N=-1000, epsilons=[0.3], n_intiaL_conditions=1)
    #     with self.assertRaises(ValueError):
    #         FHN(dt=0.1, N=1000, epsilons=[0.3], n_intiaL_conditions=0)

    #     # Test invalid parameters for Bicubic
    #     with self.assertRaises(ValueError):
    #         Bicubic(dt=-0.1, N=1000, n_intiaL_conditions=1)
    #     with self.assertRaises(ValueError):
    #         Bicubic(dt=0.1, N=-1000, n_intiaL_conditions=1)
    #     with self.assertRaises(ValueError):
    #         Bicubic(dt=0.1, N=1000, n_intiaL_conditions=0)

    #     # Test invalid parameters for GeneExpression
    #     with self.assertRaises(ValueError):
    #         GeneExpression(dt=-0.1, N=1000, n_intiaL_conditions=1)
    #     with self.assertRaises(ValueError):
    #         GeneExpression(dt=0.1, N=-1000, n_intiaL_conditions=1)
    #     with self.assertRaises(ValueError):
    #         GeneExpression(dt=0.1, N=1000, n_intiaL_conditions=0)

        # # Test invalid parameters for DelayOscillator
        # with self.assertRaises(ValueError):
        #     DelayOscillator(N=-1000)

if __name__ == '__main__':
    unittest.main()



import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm

import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader, TensorDataset

#Error handling
class NeuralNetworkError(Exception):
    """Error raised for neural network issues."""

class NeuralNetworkSetupError(NeuralNetworkError):
    """Error raised when parameters of the network are 0 but can not be zero."""
    def __init__(self, parameter_name, parameter, message="parameter has to be greater then zero/be provided."):
        self.parameter_name = parameter_name
        self.parameter = parameter
        self.message = message
        super().__init__(f'Error: {self.parameter_name} '+self.message+f' Value given: {self.parameter}')

class NeuralNetworkSetupTrainingError(NeuralNetworkError):
    """Error raised when setting up training of the neural network."""
    def __init__(self, parameter_name, parameter, message="parameter has to be greater then zero."):
        self.parameter_name = parameter_name
        self.parameter = parameter
        self.message = message
        super().__init__(f'Error: {self.parameter_name} '+self.message+f' Value given: {self.parameter}')

class NeuralNetworkDataError(NeuralNetworkError):
    """Error raised when data for the neural network is incorrect."""
    def __init__(self, errors):
        self.errors = errors
        message = "; ".join(errors)
        super().__init__(f'Error: {message}')
        

class FFNN(nn.Module):   
    """
    Feedforward Neural Network (FFNN) class.

    Args:
        nn (torch.nn module): PyTorch neural network module
    """
    def __init__(self, Nin, Nout, Nlayers, Nnodes, activation):
        super(FFNN, self).__init__()
        layers = [nn.Linear(Nin, Nnodes), activation()]
        for _ in range(Nlayers - 1):
            layers.append(nn.Linear(Nnodes, Nnodes))
            layers.append(activation())
        layers.append(nn.Linear(Nnodes, Nout))
        # layers.append(nn.Sigmoid())
        self.model = nn.Sequential(*layers)

    def forward(self, x):
        return self.model(x)

def init_weights(m):
    """
    Initialize the weights of the neural network.

    Args:
        m (torch model): PyTorch model
    """
    if isinstance(m, nn.Linear):
        nn.init.xavier_uniform_(m.weight)
        nn.init.zeros_(m.bias)

def configure_FFNN_model(Nin, Nout, Nlayers, Nnodes,  activation=nn.SiLU, optimizer_name='Adam', lr=1e-4,
                         loss_fn=nn.MSELoss, summary=False):
    """
    Configure the Feedforward Neural Network (FFNN) model.

    Args:
        Nin (int): Number of input features
        Nout (int): Number of output features
        Nlayers (int): Number of layers
        Nnodes (int): Number of nodes
        activation (torch neural network activation function module, optional): Activation function of neural network training. Defaults to nn.SiLU.
        optimizer_name (str, optional): Optimizer for model training. Defaults to 'Adam'.
        lr (float, optional): Learning rate for training. Defaults to 1e-4.
        loss_fn (torch loss function module, optional): Loss function for training. Defaults to nn.MSELoss.
        summary (bool, optional): If model summary should be generated. Defaults to False.

    Returns:
        model (torch model): setup FFNN model
        optimizer (torch optimizer): optimizer for training
        loss_fn (torch loss function): loss function for training
    """    
    if Nin == 0:
        raise NeuralNetworkSetupError('Nin', Nin)
    if Nout == 0:
        raise NeuralNetworkSetupError('Nout', Nout)
    if Nlayers == 0:
        raise NeuralNetworkSetupError('Nlayers', Nlayers)
    if Nnodes == 0:
        raise NeuralNetworkSetupError('Nnodes', Nnodes)
    if lr == 0:
        raise NeuralNetworkSetupError('Learning rate', lr)
    if optimizer_name == '':
        raise NeuralNetworkSetupError('Optimizer', optimizer_name)
    if loss_fn == '' or loss_fn == None:
        raise NeuralNetworkSetupError('Loss function', loss_fn)

    model = FFNN(Nin, Nout, Nlayers, Nnodes, activation)
    model.apply(init_weights)
    if optimizer_name == 'SGD':
        optimizer = optim.SGD(model.parameters(), lr=lr, momentum=0.9)
    elif optimizer_name == 'RMSprop':
        optimizer = optim.RMSprop(model.parameters(), lr=lr, alpha=0.99)
    elif optimizer_name == 'Adam':
        optimizer = optim.Adam(model.parameters(), lr=lr, weight_decay=1e-10)
    elif optimizer_name == 'Nadam':
        optimizer = optim.Nadam(model.parameters(), lr=lr)
    elif optimizer_name == 'Adagrad':
        optimizer = optim.Adagrad(model.parameters(), lr=lr)
    elif optimizer_name == 'Adadelta':
        optimizer = optim.Adadelta(model.parameters(), lr=lr)
    
    loss_fn = loss_fn()

    if summary:
        print(model)

    return model, optimizer, loss_fn

# Monitor gradients
def monitor_gradients(model):
    """
    Monitor the gradients of the model.

    Args:
        model (torch model feature): torch model

    Returns:
        gradients (float): gradients of the model
    """    
    for name, param in model.named_parameters():
        if param.requires_grad:
            gradients=param.grad.norm()
            return gradients

def loss_function(input, target, nc_prediction, nullcline_guess, factor):
    """
    Loss function for the neural network model.
    
    Args:
        input (torch tensor): input data
        target (torch tensor): target data
        nc_prediction (torch tensor): nullcline prediction data
        nullcline_guess (torch tensor): nullcline guess data
        factor (float): penalty factor for loss function

    Returns:
        loss (float): loss function
    """    
    mse_loss = nn.MSELoss()
    mse_loss_nc = nn.MSELoss()
    loss_train=mse_loss(input, target)
    loss_nc = mse_loss_nc(nc_prediction, nullcline_guess)
    return loss_nc+factor*loss_train

def train_FFNN_model(model, optimizer, loss_fn, input_train, target_train, input_test, target_test, validation_data,
                     epochs=200, batch_size=64, plot_loss=True, device='cpu', use_progressbar=True, save_evolution=True,
                     loss_target='limit_cycle', nullcline_guess=None, factor=1.0, method=None, minimal_value=0.0, maximal_value=1.0):
    """
    Train the Feedforward Neural Network (FFNN) model.

    Args:
        model (torch model): configured FFNN model
        optimizer (torch optimizer): optimizer
        loss_fn (torch loss function): loss function
        input_train (pandas dataframe): input training data
        target_train (pandas dataframe): target training data
        input_test (pandas dataframe): input testing data
        target_test (pandas dataframe): target testing data
        validation_data (pandas dataframe): validation data
        epochs (int, optional): number of epochs. Defaults to 200.
        batch_size (int, optional): batch size. Defaults to 64.
        plot_loss (bool, optional): plot the loss. Defaults to True.
        device (str, optional): device to train the model on. For 'cuda', necessary GPU and CUDA Toolkit are required. Defaults to 'cpu'.
        use_progressbar (bool, optional): use progress bar when running the training. Defaults to True.
        save_evolution (bool, optional): save the evolution of limit cycle and nullcline predictions. Defaults to True.
        loss_target (str, optional): Decide if the target is not limit cycle, but the inital nullcline seed. Defaults to 'limit_cycle'.
        nullcline_guess (torch tensor, optional): Torch tensor containing inital nullcline guess. Defaults to None.
        factor (float, optional): The penalty factor for nullcline guess loss evaluation. Defaults to 1.0.
        method (str, optional): Method used for prediction, can be derivative or delayed variables. Defaults to None.
        minimal_value (float, optional): Minimal value of minmax normalization for inbetween prediction. Defaults to 0.0.
        maximal_value (int, optional): Maximal value of minmax normalization for inbetween prediction. Defaults to 1.0.

    Returns:
        train_losses (list): training losses
        val_losses (list): validation losses
        test_loss (float): testing loss
        predictions (list): predictions of nullcline structure over all epochs
        lc_predictions (list): limit cycle predictions over all epochs
    """
    #Error handling
    if input_train.shape[0] == 0:
        raise NeuralNetworkDataError(['Input training data is empty.'])
    if target_train.shape[0] == 0:
        raise NeuralNetworkDataError(['Target training data is empty.'])
    if input_test.shape[0] == 0:
        raise NeuralNetworkDataError(['Input testing data is empty.'])
    if target_test.shape[0] == 0:
        raise NeuralNetworkDataError(['Target testing data is empty.'])
    
    if input_train.shape[0]!=target_train.shape[0]:
        raise NeuralNetworkDataError(['Input and target training data have different lengths.'])
    if input_test.shape[0]!=target_test.shape[0]:
        raise NeuralNetworkDataError(['Input and target testing data have different lengths.'])
    
    if validation_data[0].shape[0] == 0:
        raise NeuralNetworkDataError(['Validation input data is empty.'])
    if validation_data[1].shape[0] == 0:
        raise NeuralNetworkDataError(['Validation target data is empty.'])
    if validation_data[0].shape[0]!=validation_data[1].shape[0]:
        raise NeuralNetworkDataError(['Validation input and target data have different lengths.'])
    
    if epochs == 0:
        raise NeuralNetworkSetupTrainingError('epochs', epochs)
    if batch_size == 0:
        raise NeuralNetworkSetupTrainingError('batch_size', batch_size)     

    # Move model to the specified device
    model.to(device)
    if loss_target=='limit_cycle':
        train_dataset = TensorDataset(torch.tensor(input_train.values, dtype=torch.float32), torch.tensor(target_train.values, dtype=torch.float32))
        test_dataset = TensorDataset(torch.tensor(input_test.values, dtype=torch.float32), torch.tensor(target_test.values, dtype=torch.float32))
        val_dataset = TensorDataset(torch.tensor(validation_data[0].values, dtype=torch.float32), torch.tensor(validation_data[1].values, dtype=torch.float32))

        train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)
        test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=False)
        val_loader = DataLoader(val_dataset, batch_size=batch_size, shuffle=False)
    if loss_target=='nullcline_guess':
        
        train_dataset = TensorDataset(torch.tensor(input_train.values, dtype=torch.float32), torch.tensor(target_train.values, dtype=torch.float32), torch.tensor(np.array([nullcline_guess]*input_train.shape[0]), dtype=torch.float32))
        test_dataset = TensorDataset(torch.tensor(input_test.values, dtype=torch.float32), torch.tensor(target_test.values, dtype=torch.float32), torch.tensor(np.array([nullcline_guess]*input_test.shape[0]), dtype=torch.float32))
        val_dataset = TensorDataset(torch.tensor(validation_data[0].values, dtype=torch.float32), torch.tensor(validation_data[1].values, dtype=torch.float32), torch.tensor(([nullcline_guess]*validation_data[0].shape[0]), dtype=torch.float32))

        train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)
        test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=False)
        val_loader = DataLoader(val_dataset, batch_size=batch_size, shuffle=False)

    train_losses = []
    val_losses = []
    predictions=[]
    lc_predictions=[]
    gradients=[]

    for epoch in range(epochs):
            if use_progressbar:
                progressbar=tqdm(enumerate(train_loader), total=len(train_loader), desc=f'Epoch {epoch+1}/{epochs}')
            else:
                progressbar=enumerate(train_loader)
            model.train()
            running_loss = 0.0
            for batch_idx, data in progressbar:
                # Move inputs and targets to the specified device
                if loss_target=='limit_cycle':
                    inputs, targets = data
                    inputs, targets = inputs.to(device), targets.to(device)
                if loss_target=='nullcline_guess':
                    inputs, targets, nullcline_guess = data
                    inputs, targets, nullcline_guess = inputs.to(device), targets.to(device), nullcline_guess.to(device)

                optimizer.zero_grad()
                outputs = model(inputs)
                if loss_target=='limit_cycle':
                    loss = loss_fn(outputs, targets)

                if loss_target=='nullcline_guess':
                    input_null = np.zeros((nullcline_guess.shape[1], model.model[0].in_features))
                    for i in range(model.model[0].in_features):
                        input_null[:,i] = np.linspace(0, 1,nullcline_guess.shape[1])
                    input_null = torch.tensor(input_null, dtype=torch.float32).to(device)
                    nc_prediction = model(input_null)
                    loss = loss_function(outputs, targets, nc_prediction[:,0], nullcline_guess[0,:], factor)
                loss.backward()
                optimizer.step()
                running_loss += loss.item() * inputs.size(0)
            
            train_loss = running_loss / len(train_loader.dataset)
            train_losses.append(train_loss)

            model.eval()
            val_loss = 0.0
            with torch.no_grad():
                for data in val_loader:
                    
                    if loss_target=='limit_cycle':
                        inputs, targets = data
                        inputs, targets = inputs.to(device), targets.to(device)
                    if loss_target=='nullcline_guess':
                        inputs, targets, nullcline_guess = data
                        inputs, targets, nullcline_guess = inputs.to(device), targets.to(device), nullcline_guess.to(device)

                    outputs = model(inputs)
                    if loss_target=='limit_cycle':
                        loss = loss_fn(outputs, targets)

                    if loss_target=='nullcline_guess':
                        input_null = np.zeros((nullcline_guess.shape[1], model.model[0].in_features))
                        for i in range(model.model[0].in_features):
                            input_null[:,i] = np.linspace(0, 1,nullcline_guess.shape[1])
                        input_null = torch.tensor(input_null, dtype=torch.float32).to(device)
                        nc_prediction = model(input_null)
                        loss = loss_function(outputs, targets, nc_prediction[:,0], nullcline_guess[0,:], factor)
                    val_loss += loss.item() * inputs.size(0)

            val_loss /= len(val_loader.dataset)
            val_losses.append(val_loss)

            if save_evolution:
                _ , prediction_null = nullcline_prediction(model, Nsteps=500, method=method, min_val=minimal_value, max_val=maximal_value)
                predictions.append(prediction_null)
                input_prediction=torch.tensor(input_train.values, dtype=torch.float32)
                with torch.no_grad():
                    output_prediction = model(input_prediction).cpu().numpy()
                lc_predictions.append(output_prediction)
            if use_progressbar:
                progressbar.set_description(f'Epoch [{epoch+1}/{epochs}], Batch [{batch_idx}/{len(train_loader)}], Loss: {train_loss:.4f}')
                progressbar.refresh()  

    if plot_loss:
        # plt.plot(gradients, label='Gradients')
        plt.plot(train_losses, label='Train Loss')
        plt.plot(val_losses, label='Validation Loss')
        plt.yscale('log')
        plt.legend()
        plt.show()

    model.eval()
    test_loss = 0.0
    with torch.no_grad():
        for data in test_loader:
            if loss_target=='limit_cycle':
                inputs, targets = data
                inputs, targets = inputs.to(device), targets.to(device)
            if loss_target=='nullcline_guess':
                inputs, targets, nullcline_guess = data
                inputs, targets, nullcline_guess = inputs.to(device), targets.to(device), nullcline_guess.to(device)

            outputs = model(inputs)
            if loss_target=='limit_cycle':
                loss = loss_fn(outputs, targets)

            if loss_target=='nullcline_guess':
                input_null = np.zeros((nullcline_guess.shape[1], model.model[0].in_features))
                for i in range(model.model[0].in_features):
                    input_null[:,i] = np.linspace(0, 1,nullcline_guess.shape[1])
                input_null = torch.tensor(input_null, dtype=torch.float32).to(device)

                nc_prediction = model(input_null)
                loss = loss_function(outputs, targets, nc_prediction[:,0], nullcline_guess[0,:], factor)
            test_loss += loss.item() * inputs.size(0)

    test_loss /= len(test_loader.dataset)
    predictions_evolution=np.array(predictions)
    lc_predictions = np.array(lc_predictions)
    return train_losses, val_losses, test_loss, predictions_evolution, lc_predictions

def nullcline_prediction(model, Nsteps, device='cpu', method=None,min_val=0.0, max_val=1.0):
    """
    Predict the nullcline of the model.

    Args:
        model (torch model): FFNN model
        Nsteps (int): Number of steps for discretization of input variable for nullcline prediction
        device (str, optional): device to train the model on. For 'cuda', necessary GPU and CUDA Toolkit are required. Defaults to 'cpu'.
        method (str, optional): Method used for prediction, can be derivative or delayed variables. Defaults to None.
        min_val (float, optional): Minimum value for minmax normalization. Defaults to 0.0.  
        max_val (float, optional): Maximum value for minmax normalization. Defaults to 1.0.

    Returns:
        input_null (numpy array): input variable for nullcline prediction
        prediction_null (numpy array): predicted nullcline
    """
    input_null = np.zeros((Nsteps, model.model[0].in_features))
    if method=='derivative':
        for i in range(model.model[0].in_features-1):
            input_null[:,i] = np.linspace(min_val, max_val, Nsteps)
    else:
        for i in range(model.model[0].in_features):
            input_null[:,i] = np.linspace(min_val, max_val,Nsteps)

    input_null = torch.tensor(input_null, dtype=torch.float32).to(device)
    model.eval()
    with torch.no_grad():
        prediction_null = model(input_null).cpu().numpy()
    return input_null.cpu().numpy(), prediction_null.reshape(Nsteps,)

def compute_nullcline(fnull, xvar, yvar, Nsteps, df_coef, value_min=0.0, value_max=1.0, normalize=True):
    """
    Compute the nullcline of the model from the ground true function.

    Args:
        fnull (function): ground true function
        xvar (str): x variable for nullcline computation
        yvar (str): y variable for nullcline computation
        Nsteps (int): Number of steps for discretization of input variable for nullcline prediction
        df_coef (pandas dataframe): dataframe containing the coefficients of the model
        value_min (float, optional):  Minimum value for minmax normalization. Defaults to 0.0.
        value_max (float, optional):  Maximum value for minmax normalization.. Defaults to 1.0.
        normalize (bool, optional): If values should be minmax normalized. Defaults to True.

    Returns:
        xnull (list): x variable for nullcline prediction
        ynull (list): predicted nullcline
    """
    xnull = np.linspace(df_coef[xvar]['min'],df_coef[xvar]['max'], Nsteps)
    ynull = fnull(xnull)
    if normalize:
        ynull = (ynull - df_coef[yvar]['min'])*(value_max-value_min)/(df_coef[yvar]['max'] - df_coef[yvar]['min'])+value_min

    return xnull, ynull



import numpy as np
import pandas as pd
import os
import matplotlib.pyplot as plt
import jitcdde as jitcdde

# -------------------  FitzHugh-Nagumo oscillator -------------------
class FHN:
    """
    The Fitzhuge-Nagumo model is a simplified model of the electrical activity of a neuron. In this form taken from Prokop et al., iScience (2024).
    """    
    def __init__(self,  p=[1, 1, 0.3, 0.5, 0.0]):
        self.p = p # parameters of the model

    def model(self, U): 
        """
        Model formulation of the FHN model.
        With: 
        p = [c, d, eps, b, a]
        p = [1, 1, 0.3, 0.5, 0.0]
        Prokop et al., iScience (2024)

        Args:
            U (array): Array of initial conditions for the model.

        Returns:
            U (array): Array of the next state of the model.
        """    
        u= U[0]
        v= U[1]
        # p = [c, d, eps, b, a];   Prokop et al., iScience (2024)
        # p = [1, 1, 0.3, 0.5, 0.0]
        return np.array([-u**3 + self.p[0]*u**2 + self.p[1]*u - v,
                        self.p[2]*(u - self.p[3]*v + self.p[4])])

    def vnull(self, u):
        """
        Calculate the v-nullcline of the FHN model

        Args:
            u (array): Array of values of the v-variable.

        Returns:
            v: Array of values of the v-nullcline. 
        """
        return -u**3 + self.p[0]*u**2 + self.p[1]*u

    def unull(self, v):
        """
        Calculate the u-nullcline of the FHN model

        Args:
            v (array): Array of values of the v-variable.

        Returns:
            u: Array of values of the u-nullcline. 
        """
        return  self.p[3]*v - self.p[4]

    def fixed_points(self):
        """
        Calculate the fixed points of the FHN model.

        Returns:
            fp (Array): Array of fixed points of the model.
        """
        sol = np.roots([-self.p[3], self.p[3]*self.p[0], (self.p[1]*self.p[3]-1), -self.p[4]])

        rids = np.where(np.imag(sol)==0)
        rsol = sol[rids]

        fp = np.zeros([2,rsol.shape[0]])

        fp[0,:] = np.real(rsol)
        fp[1,:] = (fp[0,:] + self.p[4])/self.p[3]

        return fp

    def simulate(self, U, dt):
        """
        Generates the next state of the FHN model using the Runge-Kutta 4th order solver.

        Args:
            U (array): Array of initial conditions for the model.
            dt (float): Time step.

        Returns:
            U (array): Array of the next state of the model.
        """
        return rk4_solver(self.model, U, dt)
    
    def generate_data(self, x0, dt, N=10000, save=True, plot=False, max_time=2, check_period=False):
        """
        Generates synthetic data for the FHN model.

        Args:
            x0 (array): Array of initial conditions for the model.
            dt (float): Time step.
            N (int, optional): Number of time steps. Defaults to 10000.
            save (bool, optional): Save the generated data to a csv file. Defaults to True.
            plot (bool, optional): Plot the generated data. Defaults to False.
            max_time (int, optional): Maximum time for the plot. Defaults to 2.
            check_period (bool, optional): Check the period of the oscillation. Defaults to False.
        """
        df=simulate_data(self.simulate, x0, dt, N, check_period)
        if save:
            save_data(df, f'{self.__class__.__name__}_eps={self.p[2]}_a={self.p[4]}.csv')
        if plot:
            plot_data(df, max_time=max_time)


# -------------------  bicubic oscillator -------------------
class Bicubic:
    """
    The bicubic model if a model with two nonlinear nullclines (one bistable and one s-shaped) taken from Prokop et al., Chaos (2024).
    """    
    def __init__(self,p=[-0.5, 0.5, -0.3]):
        self.p = p

    def model(self, U):
        """
        Model formulation of the bicubic model.
        With: 
        p = [-0.5, 0.5, -0.3];   
        Prokop et al., Chaos (2024)

        Args:
            U (array): Array of initial conditions for the model.

        Returns:
            U (array): Array of the next state of the model.
        """  
        
        u= U[0]
        v= U[1]
        return np.array([-u**3 + u**2 + u - v,
                        self.p[0]*v**3 + self.p[1]*v**2 + self.p[2]*v + u])

    def vnull(self, u):
        """
        Calculate the v-nullcline of the bicubic model

        Args:
            u (array): Array of values of the v-variable.

        Returns:
            v: Array of values of the v-nullcline. 
        """
        return -u**3 + u**2 + u

    def unull(self, v):
        """
        Calculate the u-nullcline of the bicubic model

        Args:
            v (array): Array of values of the v-variable.

        Returns:
            u: Array of values of the u-nullcline. 
        """
        return  -(self.p[0]*v**3 + self.p[1]*v**2 + self.p[2]*v)
    
    def simulate(self, U, dt):
        """
        Generates the next state of the bicubic model using the Runge-Kutta 4th order solver.

        Args:
            U (array): Array of initial conditions for the model.
            dt (float): Time step.

        Returns:
            U (array): Array of the next state of the model.
        """
        return rk4_solver(self.model, U, dt)
    
    def generate_data(self, x0, dt, N=10000, save=True, plot=False, max_time=2, check_period=False):
        """
        Generates synthetic data for the biubic model.

        Args:
            x0 (array): Array of initial conditions for the model.
            dt (float): Time step.
            N (int, optional): Number of time steps. Defaults to 10000.
            save (bool, optional): Save the generated data to a csv file. Defaults to True.
            plot (bool, optional): Plot the generated data. Defaults to False.
            max_time (int, optional): Maximum time for the plot. Defaults to 2.
            check_period (bool, optional): Check the period of the oscillation. Defaults to False.
        """
        df=simulate_data(self.simulate, x0, dt, N, check_period)
        if save:
            save_data(df, f'{self.__class__.__name__}.csv')
        if plot:
            plot_data(df, max_time=max_time)

# -------------------  gene expression oscillator -------------------
class GeneExpression:
    """
    The gene expression model is a model of the regulation of gene expression taken from Novak & Tyson, Nature Rev Mol Cell Biol (2008).
    """    
    def __init__(self, p=[1, 0.05,  1, 0.05,   1, 0.05,  1,  1, 0.1,  2]):
        self.p = p

    def model(self, U):
        """
        Model formulation of the gene expression model.
        With: 
        p = [S,   k1, Kd,  kdx, ksy,  kdy, k2, ET,  Km, KI]
        p = [1, 0.05,  1, 0.05,   1, 0.05,  1,  1, 0.1,  2];   
        Novak & Tyson, Nature Rev Mol Cell Biol (2008)

        Args:
            U (array): Array of initial conditions for the model.

        Returns:
            U (array): Array of the next state of the model.
        """
        u= U[0]
        v= U[1]
        # p = [S,   k1, Kd,  kdx, ksy,  kdy, k2, ET,  Km, KI]
        # p = [1, 0.05,  1, 0.05,   1, 0.05,  1,  1, 0.1,  2];   Novak & Tyson, Nature Rev Mol Cell Biol (2008)
        return np.array([self.p[4]*v - self.p[5]*u - self.p[6]*self.p[7]*u/(self.p[8] + u + self.p[9]*u**2),
                        self.p[1]*self.p[0]*self.p[2]**4/(self.p[2]**4 + u**4) - self.p[3]*v])

    def vnull(self, v): 
        """
        Calculate the v-nullcline of the gene expression model

        Args:
            v (array): Array of values of the v-variable.

        Returns:
            u: Array of values of the v-nullcline. 
        """
        return self.p[1]*self.p[0]/self.p[3]*self.p[2]**4/(self.p[2]**4 + v**4)

    def unull(self, v):
        """
        Calculate the u-nullcline of the gene expression model

        Args:
            v (array): Array of values of the v-variable.

        Returns:
            u: Array of values of the u-nullcline. 
        """
        return self.p[5]*v/self.p[4] + self.p[6]*self.p[7]*v/(self.p[8] + v + self.p[9]*v**2)

    def simulate(self, U, dt):
        """
        Generates the next state of the gene expression model using the Runge-Kutta 4th order solver.

        Args:
            U (array): Array of initial conditions for the model.
            dt (float): Time step.

        Returns:
            U (array): Array of the next state of the model.
        """
        return rk4_solver(self.model, U, dt)
    
    def generate_data(self, x0, dt, N=10000, save=True, plot=False, max_time=10, check_period=False):
        """
        Generates synthetic data for the gene expression model.

        Args:
            x0 (array): Array of initial conditions for the model.
            dt (float): Time step.
            N (int, optional): Number of time steps. Defaults to 10000.
            save (bool, optional): Save the generated data to a csv file. Defaults to True.
            plot (bool, optional): Plot the generated data. Defaults to False.
            max_time (int, optional): Maximum time for the plot. Defaults to 10.
            check_period (bool, optional): Check the period of the oscillation. Defaults to False.
        """
        df=simulate_data(self.simulate, x0, dt, N, check_period)
        if save:
            save_data(df, f'{self.__class__.__name__}.csv')
        if plot:
            plot_data(df, max_time=max_time)

# -------------------  Gylcolytic oscillator -------------------
class GlycolyticOscillations:
    """
    The glycolytic oscillator is a model of glycolytic oscillations introduced in Prokop et al., iScience (2024) and in 
    this form taken from Prokop et al., Chaos (2024).
    """
    def __init__(self, p=[-0.3, -2.2, 0.25, -0.5, 0.5, 1.8, 0.7, -0.3]):
        self.p = p

    def model(self, U):
        """
        Model formulation of the glycolytic oscillations model.
        With: 
        p = [a, b, c, d, e, f, g, h]
        p = [-0.3, -2.2, 0.25, -0.5, 0.5, 1.8, 0.7, -0.3]  
        Prokop et al., Chaos (2024)

        Args:
            U (array): Array of initial conditions for the model.

        Returns:
            U (array): Array of the next state of the model.
        """
        u= U[0]
        v= U[1]
        
        return np.array([self.p[0]*u + self.p[1]*v + self.p[2]*u**2 + self.p[3]*u**3 + self.p[4]*v**3,
                        self.p[5]*u+self.p[6]*v+self.p[7]*u**3])

    def simulate(self, U, dt):
        """
        Generates the next state of the glycolytic oscillations model using the Runge-Kutta 4th order solver.

        Args:
            U (array): Array of initial conditions for the model.
            dt (float): Time step.

        Returns:
            U (array): Array of the next state of the model.
        """
        return rk4_solver(self.model, U, dt)
    
    def generate_data(self, x0, dt, N=10000, save=True, plot=False, max_time=10, check_period=False):
        """
        Generates synthetic data for the glycolytic oscillations model.

        Args:
            x0 (array): Array of initial conditions for the model.
            dt (float): Time step.
            N (int, optional): Number of time steps. Defaults to 10000.
            save (bool, optional): Save the generated data to a csv file. Defaults to True.
            plot (bool, optional): Plot the generated data. Defaults to False.
            max_time (int, optional): Maximum time for the plot. Defaults to 10.
            check_period (bool, optional): Check the period of the oscillation. Defaults to False.
        """
        df=simulate_data(self.simulate, x0, dt, N, check_period)
        if save:
            save_data(df, f'{self.__class__.__name__}.csv')
        if plot:
            plot_data(df, max_time=max_time)

# -------------------  Goodwin oscillator -------------------
class Goodwin:
    """
    The Goodwin model is a three dimensional model for phosphorylation/dephosphorylation processes of a transcription factor,
    taken in this form from Gonze et al., Acta Biotheoretica (2021), parameters from Prokop et al. (2024).
    """    
    def __init__(self, p=[1,1,1,0.1,0.1,0.1,10,1]):
        self.p = p

    def model(self, U):
        """
        Model formulation of the Goodwin model.
        With: 
        p = [a,b,c,d,e,f,n,K]
        p = [1,1,1,0.1,0.1,0.1,10,1]
        Prokop et al., iScience (2024)

        Args:
            U (array): Array of initial conditions for the model.

        Returns:
            U (array): Array of the next state of the model.
        """
        u= U[0]
        v= U[1]
        w= U[2]
        # p = [a,b,c,d,e,f,n,K]
        # p = [1,1,1,0.1,0.1,0.1,10,1];  from Prokop et al. (2024)
        return np.array([self.p[0]*(self.p[7]**self.p[6])/(self.p[7]**self.p[6] + w**self.p[6]) - self.p[3]*u,
                        self.p[1]*u - self.p[4]*v,
                        self.p[2]*v - self.p[5]*w])
    def unull(self, v, w):
        """
        Calculate the u-nullcline of the Goodwin model

        Args:
            v (array): Array of values of the v-variable.
            w (array): Array of values of the w-variable.

        Returns:
            u: Array of values of the u-nullcline. 
        """
        return self.p[0]/self.p[3]*((self.p[7]**self.p[6])/(self.p[7]**self.p[6] + w**self.p[6])) + 0 * v
    
    def vnull(self, u, w): 
        """
        Calculate the v-nullcline of the Goodwin model

        Args:
            u (array): Array of values of the v-variable.
            w (array): Array of values of the w-variable.

        Returns:
            v: Array of values of the v-nullcline. 
        """
        return self.p[1]*u/self.p[4] + 0 * w

    def wnull(self, u, v):
        """
        Calculate the w-nullcline of the Goodwin model

        Args:
            u (array): Array of values of the u-variable.
            v (array): Array of values of the v-variable.

        Returns:
            w: Array of values of the w-nullcline. 
        """
        return self.p[2]*v/self.p[5] + 0 * u

    def simulate(self, U, dt):
        """
        Generates the next state of the Goodwin model using the Runge-Kutta 4th order solver.

        Args:
            U (array): Array of initial conditions for the model.
            dt (float): Time step.

        Returns:
            U (array): Array of the next state of the model.
        """
        return rk4_solver(self.model, U, dt)
    
    def generate_data(self, x0, dt, N=10000, save=True, plot=False, max_time=10, check_period=False):
        """
        Generates synthetic data for the gene expression model.

        Args:
            x0 (array): Array of initial conditions for the model.
            dt (float): Time step.
            N (int, optional): Number of time steps. Defaults to 10000.
            save (bool, optional): Save the generated data to a csv file. Defaults to True.
            plot (bool, optional): Plot the generated data. Defaults to False.
            max_time (int, optional): Maximum time for the plot. Defaults to 10.
            check_period (bool, optional): Check the period of the oscillation. Defaults to False.
        """
        df=simulate_data(self.simulate, x0, dt, N, check_period)
        if save:
            save_data(df, f'{self.__class__.__name__}.csv')
        if plot:
            plot_data(df, max_time=max_time)

# -------------------  Oregonator oscillator -------------------

class Oregonator:
    """
    The Oregonator model is a model of the Belousov-Zhabotinsky reaction, taken from Tyson, Journal of Chemical Physics (1975).
    """
    def __init__(self, p = [0.005,3,0.60,1e-2]):
        self.p = p
    
    def model(self, U):
        """
        Model formulation of the Oregonator model.
        With: 
        p = [q,p,f,e]
        p = [0.005,3,0.60,1e-2]   
        Tyson, Journal of Chemical Physics (1975)

        Args:
            U (array): Array of initial conditions for the model.

        Returns:
            U (array): Array of the next state of the model.
        """
        x, y, z = U
        #1977 Tyson
        q=self.p[0]
        p=self.p[1]
        f=self.p[2] #0.1-2.0
        e=self.p[3]
        a=1-f+(3*f)*q/(1-f)
        b=(1-f)/(q)-(1-3*f)/(1-f)
        g=f-(f*q)/(1-f)
        d=(1-f)/(q)+(1+f)/(1-f)

        dx=(1/e)*(-a*x-b*y-q*x**2-x*y)
        dy=-g*x-d*y+f*z-x*y
        dz=(1/p)*(x-z)
        return np.array([dx, dy, dz])
      
    def xnull(self, x, z):
        """
        Calculate the x-nullcline of the Oregonator model

        Args:
            x (array): Array of values of the x-variable.
            z (array): Array of values of the z-variable.

        Returns:
            y: Array of values of the x-nullcline. 
        """
        q=self.p[0]
        p=self.p[1]
        f=self.p[2] #0.1-2.0
        e=self.p[3]
        a=1-f+(3*f)*q/(1-f)
        b=(1-f)/(q)-(1-3*f)/(1-f)
        g=f-(f*q)/(1-f)
        d=(1-f)/(q)+(1+f)/(1-f)
        return (-a*x-q*x**2)/(b+x) + 0*z
    
    def ynull(self, x, y):
        """
        Calculate the y-nullcline of the Oregonator model

        Args:
            x (array): Array of values of the x-variable.
            y (array): Array of values of the y-variable.

        Returns:
            z: Array of values of the y-nullcline. 
        """
        q=self.p[0]
        p=self.p[1]
        f=self.p[2] #0.1-2.0
        e=self.p[3]
        a=1-f+(3*f)*q/(1-f)
        b=(1-f)/(q)-(1-3*f)/(1-f)
        g=f-(f*q)/(1-f)
        d=(1-f)/(q)+(1+f)/(1-f)
        return (1/f)*(g*x+d*y+x*y)

    def znull(self, x, y):
        """
        Calculate the x-nullcline of the Oregonator model

        Args:
            x (array): Array of values of the x-variable.
            y (array): Array of values of the y-variable.

        Returns:
            z: Array of values of the z-nullcline. 
        """
        return x + 0*y
    
    def simulate(self, U, dt):
        """
        Generates the next state of the Oregonator model using the Runge-Kutta 4th order solver.

        Args:
            U (array): Array of initial conditions for the model.
            dt (float): Time step.

        Returns:
            U (array): Array of the next state of the model.
        """
        return rk4_solver(self.model, U, dt)

    def generate_data(self, x0, dt, N=10000, save=True, plot=False, max_time=10, check_period=False):
        """
        Generates synthetic data for the Oregonator model.

        Args:
            x0 (array): Array of initial conditions for the model.
            dt (float): Time step.
            N (int, optional): Number of time steps. Defaults to 10000.
            save (bool, optional): Save the generated data to a csv file. Defaults to True.
            plot (bool, optional): Plot the generated data. Defaults to False.
            max_time (int, optional): Maximum time for the plot. Defaults to 10.
            check_period (bool, optional): Check the period of the oscillation. Defaults to False.
        """
        df=simulate_data(self.simulate, x0, dt, N, check_period)
        if save:
            save_data(df, f'{self.__class__.__name__}.csv')
        if plot:
            plot_data(df, max_time=max_time)

class Lorenz:
    """
    The Lorenz model is a simple model of atmospheric convection, taken from Lorenz, Journal of the Atmospheric Sciences (1963).
    """
    def __init__(self, p = [-0.5,65]):
        self.p = p
    
    def model(self, U):
        """
        Model formulation of the Lorenz model.
        With: 
        p = [a, rho]
        p = [-0.5,65]

        Args:
            U (array): Array of initial conditions for the model.

        Returns:
            U (array): Array of the next state of the model.
        """
        x,y,z = U
        a, rho = self.p
        return np.array([a*rho*(x-y)-a*y*z, rho*x - y -x*z, -z +x*y])
    
    def xnull(self, y, z):
        """
        Calculate the x-nullcline of the Lorenz model

        Args:
            y (array): Array of values of the y-variable.
            z (array): Array of values of the z-variable.

        Returns:
            x: Array of values of the x-nullcline. 
        """
        a, rho = self.p
        return (1/rho)*y*z +y
    
    def ynull(self, x, z):
        """
        Calculate the y-nullcline of the Lorenz model

        Args:
            x (array): Array of values of the x-variable.
            z (array): Array of values of the z-variable.

        Returns:
            y: Array of values of the y-nullcline. 
        """
        a, rho = self.p
        return x*z - rho*x
    
    def znull(self, x, y):
        """
        Calculate the z-nullcline of the Lorenz model

        Args:
            x (array): Array of values of the x-variable.
            y (array): Array of values of the y-variable.

        Returns:
            z: Array of values of the z-nullcline. 
        """
        return x*y
    
    def simulate(self, U, dt):
        """
        Generates the next state of the Lorenz model using the Runge-Kutta 4th order solver.

        Args:
            U (array): Array of initial conditions for the model.
            dt (float): Time step.

        Returns:
            U (array): Array of the next state of the model.
        """
        return rk4_solver(self.model, U, dt)
    
    def generate_data(self, x0, dt, N=10000, save=True, plot=False, max_time=25, check_period=False):
        """
        Generates synthetic data for the Oregonator model.

        Args:
            x0 (array): Array of initial conditions for the model.
            dt (float): Time step.
            N (int, optional): Number of time steps. Defaults to 10000.
            save (bool, optional): Save the generated data to a csv file. Defaults to True.
            plot (bool, optional): Plot the generated data. Defaults to False.
            max_time (int, optional): Maximum time for the plot. Defaults to 25.
            check_period (bool, optional): Check the period of the oscillation. Defaults to False.
        """
        df=simulate_data(self.simulate, x0, dt, N, check_period)
        if save:
            save_data(df, f'{self.__class__.__name__}.csv')
        if plot:
            plot_data(df, max_time=max_time)

class Roessler:
    """
    The Roessler model is a simple model of a chaotic oscillator, taken from Roessler, Zeitschrift für Naturforschung A (1976).
    """
    def __init__(self, p = [0.2,0.2,1]):
        self.p = p
    
    def model(self, U):
        """
        Model formulation of the Roessler model.
        With: 
        p = [abc]
        p = [0.2,0.2,1]   
        Roessler, Zeitschrift für Naturforschung A (1976)

        Args:
            U (array): Array of initial conditions for the model.

        Returns:
            U (array): Array of the next state of the model.
        """
        x,y,z = U
        a,b,c=self.p
        dx=-y-z
        dy = x+a*y
        dz = b+(x-c)*z
        return np.array([dx, dy, dz])
    
    def xnull(self, x, z):
        """
        Calculate the x-nullcline of the Roessler model

        Args:
            x (array): Array of values of the x-variable.
            z (array): Array of values of the z-variable.

        Returns:
            y: Array of values of the x-nullcline. 
        """
        a,b,c=self.p
        return -z
    
    def ynull(self, x, z):
        """
        Calculate the y-nullcline of the Roessler model

        Args:
            x (array): Array of values of the x-variable.
            z (array): Array of values of the z-variable.

        Returns:
            y: Array of values of the y-nullcline. 
        """
        a,b,c=self.p
        return -(1/a)*x
    
    def znull(self, x, y):
        """
        Calculate the z-nullcline of the Roessler model

        Args:
            x (array): Array of values of the x-variable.
            y (array): Array of values of the y-variable.

        Returns:
            z: Array of values of the z-nullcline. 
        """
        a,b,c=self.p
        return b/(c-x)
    
    def simulate(self, U, dt):
        """
        Generates the next state of the Roessler model using the Runge-Kutta 4th order solver.

        Args:
            U (array): Array of initial conditions for the model.
            dt (float): Time step.

        Returns:
            U (array): Array of the next state of the model.
        """
        return rk4_solver(self.model, U, dt)
    
    def generate_data(self, x0, dt, N=10000, save=True, plot=False, max_time=25, check_period=False):
        """
        Generates synthetic data for the Roessler model.

        Args:
            x0 (array): Array of initial conditions for the model.
            dt (float): Time step.
            N (int, optional): Number of time steps. Defaults to 10000.
            save (bool, optional): Save the generated data to a csv file. Defaults to True.
            plot (bool, optional): Plot the generated data. Defaults to False.
            max_time (int, optional): Maximum time for the plot. Defaults to 25.
            check_period (bool, optional): Check the period of the oscillation. Defaults to False.
        """
        df=simulate_data(self.simulate, x0, dt, N, check_period)
        if save:
            save_data(df, f'{self.__class__.__name__}.csv')
        if plot:
            plot_data(df, max_time=max_time)

# -------------------  Delay oscillator -------------------
class DelayOscillator:
    """
    Simple model of a Delay Oscillator with a single delay, inspired by Lewis, Current Biology (2003)
    """    
    def __init__(self, p=[4,10,2]):
        self.p = p
    
    def model(self):
        """
        Model formulation of the delay oscillator model.
        With: 
        p = [beta, tau, n]
        p = [4,10,2]

        Args:
            U (array): Array of initial conditions for the model.

        Returns:
            U (array): Array of the next state of the model.
        """
        self.DDE = [self.p[0]/(1+jitcdde.y(0,jitcdde.t-self.p[1])**self.p[2])-jitcdde.y(0)]
        return self.DDE
    
    def unull(self, y):
        """
        Calculate the u-nullcline of the delay oscillator model

        Args:
            y (array): Array of values of the v-variable.

        Returns:
            x: Array of values of the u-nullcline. 
        """
        return self.p[0]/(1+y**self.p[2])

    def vnull(self, x):
        """
        Calculate the u-nullcline of the delay oscillator model

        Args:
            x (array): Array of values of the u-variable.

        Returns:
            y: Array of values of the v-nullcline. 
        """
        return (self.p[0]/x-1)**(1/self.p[2])
    
    def simulate(self,  dt, t_max, y_0=0):
        """
        Simulates the delay oscillatory model using the jitcdde solver.

        Args:
            dt (float): Time step.
            t_max (float): Maximum time of the simulation.
            y_0 (int, optional): Initial condition. Defaults to 0.

        Returns:
            array : Simulated data of the model.
        """
        self.model()
        DDE = jitcdde.jitcdde(self.DDE)
        DDE.constant_past(y_0)
        DDE.step_on_discontinuities()

        data = []
        for time in np.arange(DDE.t, DDE.t  +t_max, dt):
            data.append(DDE.integrate(time))
        data = np.array(data)

        return data

    def generate_data(self, y_0, dt, t_max, save=True, plot=False, check_period=False):
        """
        Generates synthetic data for the delay oscillator model.

        Args:
            y_0 (float): Initial condition for the model.
            dt (float): Time step.
            t_max (float): Maximum time of the simulation.
            save (bool, optional): Save the generated data to a csv file. Defaults to True.
            plot (bool, optional): Plot the generated data. Defaults to False.
            check_period (bool, optional): Check the period of the oscillation. Defaults to False.
        """
        data = self.simulate( dt, t_max,y_0)
        df = pd.DataFrame(data, columns=['u'])
        df['time'] = np.arange(data.shape[0])*dt
        if save:
            save_data(df, f'{self.__class__.__name__}.csv')
        if plot:
            plot_data(df, max_time=t_max)        
        
        

# -------------------  4th-order Runge-Kutta solver -------------------
    
def rk4_solver(f, u, dt):
    """
    Rung-Kutta 4th order solver for a system of ODEs.

    Args:
        f (function): Function that describes the system of ODEs. For model classes it is the model method.
        u (float): Value of a state variable at time t.
        dt (float): Time step.

    Returns:
        u[i+1] (float): Value of a state variable at time t+dt.
    """    
    k1 = f(u)
    k2 = f(u + k1*dt/2)
    k3 = f(u + k2*dt/2)
    k4 = f(u + k3*dt)

    return u + dt/6*(k1 + 2*k2 + 2*k3 + k4)

# -------------------  Generate and save data of multiple IC to csv -------------------

def simulate_data (simulate, x0, dt, N=10000, check_period=False):
    """
    Generate synthetic data for a given model and initial conditions.

    Args:
        simulate (method): Method that simulates the model. For model classes it is the simulate method.
        x0 (array of floats): Initial conditions for the model.
        dt (float): Time step.
        N (int, optional): Number of timesteps. Defaults to 10000.
        check_period (bool, optional): Limits the amount of periods calculated. Defaults to True.

    Returns:
        df (pandas DataFrame): DataFrame with the synthetic data.
    """    
    df = pd.DataFrame()
    sim_count = 1
    time = np.arange(N)*dt
    # print(x0,len(x0), N)
    u = np.zeros((len(x0),N))
    if len(x0)==2:
        for i in range(x0[0,:,:].shape[0]):
            for j in range(x0[0,:,:].shape[1]):
                
                u[:,0] = [x0[0,i,j], x0[1,i,j]]
                

                for n in range(1,N):
                    u[:,n] = simulate(u[:,n-1], dt)
                
                df_sim = pd.DataFrame()

                df_sim['sim'] = np.ones(u.shape[1])*sim_count
                df_sim['time'] = time
                var_name=['u', 'v', 'w', 'x', 'y']
                for i_var in range(len(x0)):
                    df_sim[f'{var_name[i_var]}'] = u[i_var,:]
                
                
        
                if check_period:
                    _,_,period,_=calculate_period(df_sim['u'].to_numpy(), df_sim['time'].to_numpy())
                    if period is not None:
                        max_time = 6 * period
                        df_sim = df_sim[df_sim['time'] <= max_time]

                sim_count += 1

                df = pd.concat((df, df_sim), ignore_index=True)
    if len(x0)==3:
        for i in range(x0[0,:,:,:].shape[0]):
            for j in range(x0[0,:,:,:].shape[1]):
                for k in range(x0[0,:,:,:].shape[2]):
                    u[:,0] = [x0[0,i,j,k], x0[1,i,j,k], x0[2,i,j,k]]

                    for n in range(1,N):
                        u[:,n] = simulate(u[:,n-1], dt)
                    
                    df_sim = pd.DataFrame()

                    df_sim['sim'] = np.ones(u.shape[1])*sim_count
                    df_sim['time'] = time
                    var_name=['u', 'v', 'w']
                    for i_var in range(len(x0)):
                        df_sim[f'{var_name[i_var]}'] = u[i_var,:]
                    
                    if check_period:
                        _,_,period,_=calculate_period(df_sim['u'].to_numpy(), df_sim['time'].to_numpy())
                        if period is not None:
                            max_time = 6 * period
                            df_sim = df_sim[df_sim['time'] <= max_time]

                    sim_count += 1

                    df = pd.concat((df, df_sim), ignore_index=True)

    return df

def save_data(df, filename):
    """
    Save the generated data to a csv file.

    Args:
        df (pandas DataFrame): DataFrame with the synthetic data.
        filename (str): Name of the file, usually taking the name of the model.
    """    
    # Create 'data' directory if it does not exist
    if not os.path.exists('data/synthetic_data'):
        os.makedirs('data/synthetic_data')

    # Save the DataFrame to the 'data' directory
    filepath = os.path.join('data/synthetic_data', filename)
    
    # Save the DataFrame to the 'data/synthetic_data' directory
    df.to_csv(filepath, index=False)
    print('Generated data saved to', filepath)


def plot_data(df, max_time):
    """
    Plot the generated data.

    Args:
        df (pandas DataFrame): DataFrame with the synthetic data.
        max_time (float): Maximum time for the plot.
    """    
    fig, ax = plt.subplots(1,1)

    for i in df['sim'].unique():
        df_sim = df[(df['sim']==i) & (df['time']<=max_time)].copy()
        df_sim.plot.line(x='u', y='v', ax=ax, legend=False)

    ax.set_ylabel('v')


from scipy.signal import find_peaks

#### General functions
def calculate_period(x_train, t_train):
    """
    Calculate the period of an oscillation.

    Parameters
    ----------
    x_train : np.array
        Time series of data.
    t_train : np.array
        Time of the time series.

    Returns
    -------
    peaks_u : np.array
        Size of peaks detected in the time series.
    peaks_t : np.array
        Time points where peaks occur in the data set.
    period : float
        Period of Time Series.
    peaks[0]: list
        Array of all indicies of local maxima
    
    """
    if len(x_train.shape)<2:
        peaks=find_peaks(x_train)
        peaks_t=t_train[peaks[0]]
        peaks_u=x_train[peaks[0]]
    elif x_train.shape[1]==2 or x_train.shape[1]==3:
        peaks=find_peaks(x_train[:,0])
        peaks_t=t_train[peaks[0]]
        peaks_u=x_train[peaks[0],0]
    elif x_train.shape[1]==5:
        peaks=find_peaks(x_train[:,1])
        peaks_t=t_train[peaks[0]]
        peaks_u=x_train[peaks[0],1]

    period_temp=0
    for i in range(len(peaks_t)-1):
        period_temp=period_temp+(peaks_t[i+1]-peaks_t[i])
    if len(peaks_t)>1:
        subtract=1
    else: subtract=0
    period=period_temp/(len(peaks_t)-subtract)
    return peaks_u, peaks_t, period, peaks[0]


import numpy as np
from . import model

def FHN(dt=0.1, N=1000000, epsilons=[0.3], n_intiaL_conditions=1):
    """
    Generate data using the FitzHugh-Nagumo model (see src/pyCLINE/model.py)
    with different time scale separation, as used in Prokop, Billen, Frolov, Gelens (2025).

    Args:
        dt (float): Time step. Defaults to 0.1.
        N (int): Number of time steps. Defaults to 1000000.
        epsilons (list): List of time scale separations. Defaults to [0.3].
    
    Returns:
        None
    """
    if dt <= 0:
        raise ValueError("Time step (dt) must be positive.")
    if N <= 0:
        raise ValueError("Number of time steps (N) must be positive.")
    if n_intiaL_conditions <= 0:
        raise ValueError("Number of initial conditions must be positive.")

    u = np.zeros([2,N])
    for i_eps, eps in enumerate(epsilons):
        u[:,0] = [0.1, 0.1]
        p = [1, 1, eps, 0.5, 0.0]
        fhn=model.FHN(p)
        u0, v0 = np.meshgrid(np.linspace(-1.25,1.75,n_intiaL_conditions),np.linspace(-0.75,1.75,n_intiaL_conditions))
        x0 = np.array([u0,v0])
        fhn.generate_data(x0, dt, N)
    pass

def Bicubic(dt=0.1, N=1000000, n_intiaL_conditions=1):
    """
    Generate data using the Bicubic model (see src/pyCLINE/model.py),
    as used in Prokop, Billen, Frolov, Gelens (2025).

    Args:
        dt (float, optional): Time step. Defaults to 0.1.
        N (int, optional): Number of time steps. Defaults to 1000000.
    """
    if dt <= 0:
        raise ValueError("Time step (dt) must be positive.")
    if N <= 0:
        raise ValueError("Number of time steps (N) must be positive.")
    if n_intiaL_conditions <= 0:
        raise ValueError("Number of initial conditions must be positive.")

    u = np.zeros([2,N])

    u[:,0] = [0.1, 0.1]
    p = [-0.5, 0.5, -1/3]

    bicubic=model.Bicubic(p)
    u0, v0 = np.meshgrid(np.linspace(-1.25,1.75,n_intiaL_conditions),np.linspace(-0.75,1.75,n_intiaL_conditions))
    x0 = np.array([u0,v0])
    bicubic.generate_data(x0, dt, 10000)
    pass

def GeneExpression(dt=0.1, N=1000000, n_intiaL_conditions=1):
    """
    Generate data using the Gene Expression model (see src/pyCLINE/model.py),
    as used in Prokop, Billen, Frolov, Gelens (2025).

    Args:
        dt (float, optional): Time step. Defaults to 0.1.
        N (int, optional): Number of time steps. Defaults to 1000000.
    """
    if dt <= 0:
        raise ValueError("Time step (dt) must be positive.")
    if N <= 0:
        raise ValueError("Number of time steps (N) must be positive.")
    if n_intiaL_conditions <= 0:
        raise ValueError("Number of initial conditions must be positive.")

    u = np.zeros([2,N])

    u[:,0] = [0.1, 0.1]
    p=[1, 0.05,  1, 0.05,   1, 0.05,  1,  1, 0.1,  2]

    gene_expression=model.GeneExpression(p)
    u0, v0 = np.meshgrid(np.linspace(0,1.75,n_intiaL_conditions),np.linspace(0,1.75,n_intiaL_conditions))
    x0 = np.array([u0,v0])
    gene_expression.generate_data(x0, dt, 10000)
    pass

def DelayOscillator(N=20000):
    """
    Generate data using the Delay Oscillator model (see src/pyCLINE/model.py),
    as used in Prokop, Billen, Frolov, Gelens (2025).

    Args:
        N (int, optional): Number of time steps. Defaults to 20000.
    """
    if N <= 0:
        raise ValueError("Number of time steps (N) must be positive.")
    
    time = np.linspace(0, 400, N-1)
    dt=time[1]-time[0]
    p=[4, 10, 2]

    delay_osci=model.DelayOscillator(p)
    delay_osci.generate_data( y_0=0, dt=dt, t_max=time[-1])
    pass